"""Includes functionality for inline editing of SOI."""
import string
from datetime import datetime
from PySide2.QtCore import Qt, QRectF, QTimer, QPoint, QMarginsF
from PySide2.QtWidgets import (
    QApplication,
    QScrollArea,
    QLabel,
    QGraphicsScene,
    QGraphicsView,
    QGraphicsRectItem,
)
from PySide2.QtGui import QFont, QPixmap, QBrush, QPalette, QPainter
from PySide2.QtPrintSupport import QPrinter
from soitool.soi import ModuleLargerThanBinError
from soitool.dialog_wrappers import exec_warning_dialog
from soitool.serialize_export_import_soi import generate_soi_filename

# How attachment pages should be numbered. The first page should be numbered
# by the value of ATTACHMENT_NUMBERING_SCHEME[0], the second page
# ATTACHMENT_NUMBERING_SCHEME[1], and so on.
ATTACHMENT_NUMBERING_SCHEME = list(string.ascii_uppercase)


class InlineEditableSOIView(QScrollArea):
    """Widget som kan byttes ut med view, edit etc."""

    def is_widget_in_scene(self, widget):
        """Indicate wether given widget already has a proxy in the scene."""
        for proxy in self.proxies:
            if proxy.widget() == widget:
                return True
        return False

    def ensure_proxies(self):
        """Make sure all modules of the SOI have a proxy inside the scene."""
        for module in self.soi.modules + self.soi.attachments:
            if not self.is_widget_in_scene(module["widget"]):
                proxy = self.scene.addWidget(module["widget"])
                self.proxies.add(proxy)

    def mousePressEvent(self, _):
        """Reorganize modules when pressed.

        This is a temporary way to activate reorganization of widgets. Note
        that will not be triggered by clicks on modules in the scene.

        If reorganization cannot occur because a module is too large the user
        will be informed. It's worth noting that with the current
        implementation this is the only way the user will be informed of this,
        so reorganization that is not triggered here will not give feedback to
        the user.
        """
        try:
            self.soi.reorganize()
        except ModuleLargerThanBinError:
            exec_warning_dialog(
                text="Minst én av modulene er for store "
                "for å passe på én side!",
                informative_text="Programmet kan desverre ikke fikse dette "
                "for deg. Se igjennom modulene og sjekk at "
                "alle moduler er mindre enn én enkelt side.",
            )

    def __init__(self, soi):
        super().__init__()

        self.soi = soi

        self.number_of_pages_total = 1
        self.number_of_non_attachment_pages = 0
        self.proxies = set()

        # Necessary to make the scroll area fill the space it's given
        self.setWidgetResizable(True)

        # Scene and view widgets
        self.scene = QGraphicsScene()
        self.setup_scene()
        self.view = QGraphicsView(self.scene)

        self.setWidget(self.view)

        self.ensure_proxies()
        self.update_pages()

        # Add listeners to react properly to SOI changes
        self.soi.add_reorganization_listener(self.update_pages)
        self.soi.add_new_module_listener(self.ensure_proxies)
        self.soi.add_update_property_listener(self.update_pages)

        # self.launch_auto_zoom()

    def produce_pdf(self, filename=None):
        """Produce PDF using QGraphicsScene.

        Renders the QGraphicsScene-representation of the SOI as a PDF. This
        PDF is in theory supposed to be searchable [1], but in practice it
        seems that each QGraphicsItem in the scene is simply dumped as an
        image.

        Sources:
        [1]: https://doc.qt.io/qt-5/qprinter.html#OutputFormat-enum

        Parameters
        ----------
        filename : str
            Name of file to store PDF in. If file exists it will be
            overwritten. Note that the filename should contain the extension
            '.pdf' to be properly handled by operating systems. If no filename
            is supplied one will be generated following the same convention as
            for exported JSON and TXT files

        Raises
        ------
        ValueError
            If filename is invalid.
        """
        if filename is None:
            filename = generate_soi_filename(self.soi) + ".pdf"

        printer = QPrinter(QPrinter.HighResolution)
        printer.setOutputFormat(QPrinter.PdfFormat)
        printer.setOutputFileName(filename)
        printer.setPageSize(QPrinter.A4)
        printer.setOrientation(QPrinter.Landscape)
        printer.setPageMargins(QMarginsF(0, 0, 0, 0))

        painter = QPainter()

        try:
            ok = painter.begin(printer)

            if not ok:
                raise ValueError(
                    "Not able to begin QPainter using QPrinter "
                    "based on argument "
                    "filename '{}'".format(filename)
                )

            # Render each page to own PDF page
            for i in range(self.number_of_pages_total):

                x = 0
                y = self.soi.HEIGHT * i + self.soi.PADDING * i

                self.scene.render(
                    painter,
                    source=QRectF(x, y, self.soi.WIDTH, self.soi.HEIGHT),
                )

                # If there are more pages, newPage
                if i + 1 < self.number_of_pages_total:
                    printer.newPage()
        finally:
            painter.end()

    def update_number_of_pages(self):
        """Make sure number of pages excactly fits SOI modules.

        The minimum page count is 1.
        """
        self.number_of_non_attachment_pages = (
            self.soi.get_number_of_non_attachment_pages()
        )
        # each attachment module requires it's own page
        required_pages = self.number_of_non_attachment_pages + len(
            self.soi.attachments
        )
        self.number_of_pages_total = required_pages

    def draw_pages(self):
        """Draw self.number_of_pages_total pages."""
        for i in range(self.number_of_pages_total):

            x = 0
            y = self.soi.HEIGHT * i + self.soi.PADDING * i

            # Adjust page size
            full_scene_height = y + self.soi.HEIGHT
            self.scene.setSceneRect(
                QRectF(0, 0, self.soi.WIDTH, full_scene_height)
            )

            self.draw_page(x, y)
            page_number = i + 1
            header_x = x + self.soi.PADDING
            header_y = y + self.soi.PADDING
            # if not an attachment page: draw as normal
            # if attachment page: draw starting from 1 again
            if page_number <= self.number_of_non_attachment_pages:
                self.draw_header(header_x, header_y, page_number, False)
            else:
                self.draw_header(
                    header_x,
                    header_y,
                    page_number - self.number_of_non_attachment_pages,
                    True,
                )

        for proxy in self.proxies:
            # Redraw of pages requires modules to be moved to front again
            proxy.setZValue(1)

    def update_pages(self):
        """Update pages drawn in the scene to reflect the SOI."""
        self.update_number_of_pages()
        self.draw_pages()

    # Ignoring Pylint's errors "Too many local variables" and
    # "Too many statements" for this function because it's doing GUI layout
    # work, which in nature is tedious and repetitive.
    # pylint: disable=R0914,R0915
    def draw_header(self, x, y, page_number, is_attachment_page):
        """Draw header staring at given position.

        Source for rotation approach:
        * https://stackoverflow.com/a/43389394/3545896

        Parameters
        ----------
        x : int
        y : int
        page_number : int
            Page number of page to draw. 'is_attachment_page' affects how the
            page number is drawn.
        is_attachment_page : bool
            If True the page will indicate that it is an attachment, and the
            page numbering will follow a convention for page numbering defined
            in 'ATTACHMENT_NUMBERING_SCHEME'. If False the page will indicate
            that the page is part of the main page, and the page numbering will
            use 1,2,3,....
        """
        # Title
        label_title = QLabel(self.soi.title)
        label_title.move(x, y + self.soi.HEADER_HEIGHT)
        label_title.setStyleSheet("background-color: rgba(0,0,0,0%)")
        label_title.setFont(QFont("Times New Roman", 35))
        proxy = self.scene.addWidget(label_title)
        proxy.setRotation(-90)

        # Description
        label_description = QLabel(self.soi.description)
        label_description.move(x + 57, y + self.soi.HEADER_HEIGHT)
        label_description.setStyleSheet("background-color: rgba(0,0,0,0%)")
        label_description.setFont(QFont("Times New Roman", 15))
        proxy = self.scene.addWidget(label_description)
        proxy.setRotation(-90)

        # Creation date
        creation_date = soi_date_string_to_user_friendly_string(self.soi.date)
        label_creation_date = QLabel("Opprettet: {}".format(creation_date))
        label_creation_date.setStyleSheet("background-color: rgba(0,0,0,0%)")
        label_creation_date.setFont(QFont("Times New Roman", 15))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            label_creation_date.fontMetrics()
            .boundingRect(label_creation_date.text())
            .width()
        )
        label_creation_date.move(x + 80, y + self.soi.HEADER_HEIGHT)
        proxy = self.scene.addWidget(label_creation_date)
        proxy.setRotation(-90)

        # Patch
        pixmap = QPixmap(self.soi.icon)
        patch = QLabel()
        patch.setPixmap(
            pixmap.scaled(self.soi.HEADER_WIDTH - 2, self.soi.HEADER_WIDTH - 2)
        )
        patch.move(
            x + 1, y + self.soi.HEADER_HEIGHT / 2 + self.soi.HEADER_WIDTH + 10
        )
        proxy = self.scene.addWidget(patch)
        proxy.setRotation(-90)

        # Copy number
        copy_number = QLabel("1 av N")
        copy_number.setStyleSheet("background-color: rgba(0,0,0,0%)")
        copy_number.setFont(QFont("Times New Roman", 40))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            copy_number.fontMetrics().boundingRect(copy_number.text()).width()
        )
        # Store for usage when placing copy number title and page number
        copy_number_y_pos = y + self.soi.HEADER_HEIGHT / 2 - 100
        copy_number.move(x + 15, copy_number_y_pos + label_width / 2)
        proxy = self.scene.addWidget(copy_number)
        proxy.setRotation(-90)

        # Copy number title
        label_copy_number_header = QLabel("Eksemplar")
        label_copy_number_header.setStyleSheet(
            "background-color: rgba(0,0,0,0%)"
        )
        label_copy_number_header.setFont(QFont("Times New Roman", 15))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            label_copy_number_header.fontMetrics()
            .boundingRect(label_copy_number_header.text())
            .width()
        )
        label_copy_number_header.move(x, copy_number_y_pos + label_width / 2)
        proxy = self.scene.addWidget(label_copy_number_header)
        proxy.setRotation(-90)

        # Page numbering
        if is_attachment_page:
            page_number = QLabel(
                "Vedlegg {}".format(
                    ATTACHMENT_NUMBERING_SCHEME[page_number - 1]
                )
            )
        else:
            page_number = QLabel(
                "Side {} av {}".format(
                    page_number, self.number_of_non_attachment_pages
                )
            )
        page_number.setStyleSheet("background-color: rgba(0,0,0,0%)")
        page_number.setFont(QFont("Times New Roman", 15))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            page_number.fontMetrics().boundingRect(page_number.text()).width()
        )
        page_number.move(x + 80, copy_number_y_pos + label_width / 2)
        proxy = self.scene.addWidget(page_number)
        proxy.setRotation(-90)

        # Classification
        classification = QLabel(self.soi.classification)
        classification.setStyleSheet(
            "background-color: rgba(0,0,0,0%); " "color: red"
        )
        classification.setFont(QFont("Times New Roman", 35))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            classification.fontMetrics()
            .boundingRect(classification.text())
            .width()
        )
        classification.move(x, y + label_width)
        proxy = self.scene.addWidget(classification)
        proxy.setRotation(-90)

        # From date
        valid_from = soi_date_string_to_user_friendly_string(
            self.soi.valid_from
        )
        label_valid_from = QLabel("Gyldig fra: {}".format(valid_from))
        label_valid_from.setStyleSheet("background-color: rgba(0,0,0,0%)")
        label_valid_from.setFont(QFont("Times New Roman", 15))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            label_valid_from.fontMetrics()
            .boundingRect(label_valid_from.text())
            .width()
        )
        label_valid_from.move(x + 57, y + label_width)
        proxy = self.scene.addWidget(label_valid_from)
        proxy.setRotation(-90)

        # To date
        valid_to = soi_date_string_to_user_friendly_string(self.soi.valid_to)
        label_valid_to = QLabel("Gyldig til: {}".format(valid_to))
        label_valid_to.setStyleSheet("background-color: rgba(0,0,0,0%)")
        label_valid_to.setFont(QFont("Times New Roman", 15))
        # Source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            label_valid_to.fontMetrics()
            .boundingRect(label_valid_to.text())
            .width()
        )
        label_valid_to.move(x + 80, y + label_width)
        proxy = self.scene.addWidget(label_valid_to)
        proxy.setRotation(-90)

    def draw_page(self, x, y):
        """Draw page starting at given position.

        Parameters
        ----------
        x : int
        y : int
        """
        # Color the page white
        page_background = QGraphicsRectItem(
            x, y, self.soi.WIDTH, self.soi.HEIGHT
        )
        page_background.setBrush(QBrush(Qt.white))
        self.scene.addItem(page_background)

        # Draw borders
        self.scene.addRect(x, y, self.soi.WIDTH, self.soi.HEIGHT)
        self.scene.addRect(
            x + self.soi.PADDING,
            y + self.soi.PADDING,
            self.soi.WIDTH - self.soi.PADDING * 2,
            self.soi.HEIGHT - self.soi.PADDING * 2,
        )
        self.scene.addRect(
            x + self.soi.PADDING,
            y + self.soi.PADDING,
            self.soi.HEADER_WIDTH,
            self.soi.HEADER_HEIGHT,
        )

    def setup_scene(self):
        """Prepare scene for use.

        Draws borders, background, etc.
        """
        # Sets the background color of the scene to be the appropriate
        # system-specific background color
        # source: https://stackoverflow.com/a/23880531/3545896
        # source: https://stackoverflow.com/questions/15519749/how-to-get-widget-background-qcolor
        app = QApplication.instance()
        self.scene.setBackgroundBrush(app.palette().color(QPalette.Window))

        self.update_pages()

    def launch_auto_zoom(self):
        """Zoom in a regular interval.

        Used to demonstrate zooming only, should be removed once the project
        matures.
        """

        def do_on_timeout():
            self.zoom(1 / 1.00005)

        timer = QTimer(self)
        timer.timeout.connect(do_on_timeout)
        timer.start(10)

    def zoom(self, zoom_factor):
        """Zoom GraphicsView by zoom_factor.

        source: https://stackoverflow.com/a/29026916/3545896

        Parameters
        ----------
        zoom_factor : float
            > 1 to zoom in and < 1 to zoom out
        """
        # Set Anchors
        self.view.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.view.setResizeAnchor(QGraphicsView.NoAnchor)

        # Zoom
        self.view.scale(zoom_factor, zoom_factor)

        pos_old = self.view.mapToScene(
            QPoint(self.soi.WIDTH / 2, self.soi.HEIGHT / 2)
        )
        pos_new = self.view.mapToScene(self.soi.WIDTH / 2, self.soi.HEIGHT / 2)
        delta = pos_new - pos_old
        self.view.translate(delta.x(), delta.y())


def soi_date_string_to_user_friendly_string(date_string):
    """Convert SOI date to user-friendly format.

    Parameters
    ----------
    date_string : str
        String following convention used in SOI, which is 'YYYY-MM-DD'

    Returns
    -------
    str
        String following user friendly convention 'DD.MM.YYYY'
    """
    parsed_date = datetime.strptime(date_string, "%Y-%m-%d")
    return parsed_date.strftime("%d.%m.%Y")
